/************************************************************************
 *                      GENESIS.C                                       *
 *                                                                      *
 *                         by Kaneda                                    *
 *                                                                      *
 *                                                                      *
 * History:                                                             *
 *      ?????98         First release                                   *
 *      20Nov98         Make it more readable :)                        *
 *      26Dec98         Put OUTTEXT_XY in CONIO.C                       *
 *      09Mar99         Add WAIT_PRESS (useful, no?)                    *
 *      16Aug99         Add Flipping in SHOW_TILES                      *
 *      31Aug99         Work on Scroll functions to add field B         *
 *                      Convert ScrollA functions in ScrollAB           *
 *	24Feb01		Correct the awful sprite bug!															*
 *			I don't know how my others demos were able to										*
 *				work with a so awful bug!																			*
 *	24May02		RAZ now correctly raz scrollV plane B											*
 *	22Mar04		Some improvments																					*
 *  02Mar05		Time to clean it a little ;)															*
 *                                                                      *
 ************************************************************************/


/****** TO DO ******
 *
 * - rename all fucntions as Gen_xxxxxx
 * - same way to handle param (type param1, type param2...)
 * - check valid data in DMA VRAM COPY
 * - check valid data in show_tiles
 * - check valid data in show_sprites
 * - modify wait_press( ) to wait_press(key)
 * 
 ********************/
#include "genesis.h"

 #include "debug.h"
/*************    Variables  *************/

ushort  KANGEN_sprites_def[40*5]; /* 40 sprites max */
static   ushort  KANGEN_pal_black[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}; /* black palette */
volatile ushort  extWIDTH;
volatile ushort  extHEIGHT;



/************* SCREEN_SIZE ************
 * Define size of the screen, in tiles
 * This size is used everywhere so
 * you MUST call this func!
 **************************************/
void Gen_ScreenSize( ushort width, ushort height)
{
	extWIDTH = width;
	extHEIGHT = height;	
}


/************* DMA_VRAM_COPY ************
 * Use the DMA to copy the map data to  *
 * the GFX RAM.                         *
 * source = array of ulong              *
 * dest   = first number of the tiles   *
 *         (32bytes*existing_tiles)     *
 * len    = length of ulong             *
 *         (32bytes*nbr_tiles)          *
 *                                      *
 *   Made by P.Lee                      *
 ****************************************/
void dma_vram_copy(uint source,ushort dest,ushort len)
{
    register volatile ushort *pw;
    register uint *pl;

    /* Halt the Z80 for DMA */
    pw = (ushort *) Z80_HALT;
    *pw = 0x100;

    /* Point to GFX register */
    pw = (ushort *) GFXCNTL;

    /* Setup DMA length */
    len >>= 1;
    *pw = 0x9300L + (len & 0xff);
    *pw = 0x9400L + ((len >> 8) & 0xff);

    /* Setup DMA address */
    source >>= 1;
    *pw = 0x9500L + (source & 0xff);

    source >>= 8;
    *pw = 0x9600L + (source & 0xff);

    source >>= 8;
    *pw = 0x9700L + (source & 0xff);

    /* Enable DMA */
    pl = (uint *) GFXCNTL;
    *pl = GFX_DMA_ADDR((uint) dest);

    /* Enable Z80 */
    pw = (ushort *) Z80_HALT;
    *pw = 0;
}


/**************** SHOW_TILES ********************
 * Show the tiles by writing the tile           *
 * number (1 to...) to the GFX memory.          *
 *     priority flipV           tile number     *
 *            |  |  _____________|              *
 * TileDef =  0000 0000 0000 0000               *
 *             ||  |                            *
 *        palette  flipH                        *
 *                                              *
 * fieldAdr = adress of plan of dest            *
 * pri = priority                               *
 *                                              *
 *      Made by P.Lee                           *
 *      Modified by Kaneda to allow arg         *
 *              allow Flipping                  *
 *              assume you define the WIDTH     *
 ************************************************/
void show_tiles(ushort from, ushort to, uchar pal, ushort posx, ushort posy,\
	 ushort fieldAdr, uchar pri, uchar flipV, uchar flipH)
{
	volatile uint *pl;
	volatile ushort *pw;
	register  uint i;
	register uint TileDef;

	/* Point to the coordinate we want (WIDTH=Plane width defined in .H) */
	i = fieldAdr + (  2 * (posx + (extWIDTH*posy) ) ); 
	
	/* We make TileDef (see up)*/
	TileDef = pri&0x01;
 	TileDef <<= 2;
	TileDef |= pal&0x03;
	TileDef <<= 1;
	TileDef |= flipV&0x01;
	TileDef <<= 1;
	TileDef |= flipH&0x01;
	TileDef <<= 11;

	pl = (uint *) GFXCNTL;
	*pl = GFX_WRITE_ADDR(i);

	pw = (ushort *) GFXDATA;
	for (i=from;i < to+1;i++)
		*pw = (TileDef | i);
}


/**************** SHOW_UNIQUE *******************
 * Show a by writing the tile property    		*
 *     priority flipV           tile number     *
 *            |  |  _____________|              *
 * TileDef =  0000 0000 0000 0000               *
 *             ||  |                            *
 *        palette  flipH                        *
 *                                              *
 * fieldAdr = adress of plan of dest            *
 * pri = priority                               *
 *                                              *
 *      Made by Kaneda							*
 *              assume you define the WIDTH     *
 ************************************************/
void Gen_ShowUnique(ushort posx, ushort posy, ushort tileDef, ushort fieldAdr) //, uchar pal)
{
	volatile uint *pl;
	volatile ushort *pw;
	register uint i;
//	register uint fullDef;

	/* Point to the coordinate we want */
	i = fieldAdr + (  2 * (posx + (extWIDTH*posy) ) ); 
	
	pl = (uint *) GFXCNTL;
	*pl = GFX_WRITE_ADDR(i);

//	fullDef = (ushort) pal;
//	fullDef <<= 13;
//	fullDef |= tileDef;

	pw = (ushort *) GFXDATA;
//	*pw = fullDef;
	*pw = tileDef;
}

/***************** DEF_SPRITE *******************
 * Definition of the sprite                     *
 * number = number of the sprite                *
 * posx, posy =position of the sprite           *
 * size = in hexa                               *
 *        ex: 0xb00 = 3*4 (b=1011)              *
 * from =                                       *
 *      bit 15= priority                        *          
 *      bit 14-13 = palette                     *          
 *      bit 12 = vertical flipping              *          
 *      bit 11 = horizontal flipping            *          
 *      bit 10-0 = tile number                  *
 *                                              *
 *      Made by Kaneda                          *
 ************************************************/
BOOL Gen_DefineSprite(uchar number,ushort posx,ushort posy,ushort size,ushort from)
{
        /* As you can see, all is based on a uint array         */
        /* It's my choice, but you can find something else      */
        ushort loc;
        
        if ( number >= 40 )	return FALSE;
        
        loc=5*number;
        KANGEN_sprites_def[loc]=posy&0x03FF;
        KANGEN_sprites_def[loc+1]=size&0x0F00;
        KANGEN_sprites_def[loc+2]=number&0x007F;
        KANGEN_sprites_def[loc+3]=from&0xFFFF;
        KANGEN_sprites_def[loc+4]=posx&0x03FF;
        
        return TRUE;
}

/*********************** SHOW_SPRITE ********************
 *  Show the sprites                                    *
 *  from = number of the first sprite to show           *
 *  to   =   "    "   "  last    "    "   "             *
 *                                                      *
 *    Made by P.Lee                                     *
 *    Modified by Kaneda to use SPRITES_DEF[]           *
 ********************************************************/
void Gen_ShowSprites(uchar from, uchar to)
{
    register uint *pl;
    register ushort *pw;
    register ushort i;

    /* Sprite table begins at (see GENESIS.H)  */
    /* Note: We set this address using         */
    /*       reg. 85 in the init_gfx routine. */
    pl = (uint *) GFXCNTL;
    *pl = GFX_WRITE_ADDR(SLIST);

    pw = (ushort *) GFXDATA;    

    for ( i = from ; i < to ; i++ )
    {
            /* Set the y position */
            *pw = KANGEN_sprites_def[5*i];
        
            /* We need to set the size and number of the */
            /* sprite we want to show.  0=last, 1=first, */
            /* 2=second, etc...                          */
            /* 0x500 = 2x2 tile sprite                   */
            /* 0 = last sprite                           */
            if (i == (to-1))
            	*pw = KANGEN_sprites_def[5*i +1]; 
            else
            	*pw = (KANGEN_sprites_def[5*i +1] | (i+1)); 
        
            /* bit 15= priority                          */
            /* bit 14-13 = palette                       */
            /* bit 12 = vertical flipping                */
            /* bit 11 = horizontal flipping              */
            /* bit 10-0 = tile number                    */
            /* The genesis shows sprites x then y.       */
            /* Note: Look at the tile data.              */
            *pw = KANGEN_sprites_def[5*i +3];
        
            /* x position */
            *pw = KANGEN_sprites_def[5*i +4];
    }
}

/********************* RESET_SPRITES ********************
 * Reset of the nbr first sprites                       *
 *                                                      *
 *  Made by Kaneda                                      *
 *	0.1	correct a big error!			*
 ********************************************************/
void Gen_ResetSprites( uchar nbr)
{
        register uint *pl;
        register volatile ushort *pw;
        ushort i;

		nbr*=5;
		
        pl = (uint *) GFXCNTL;
        *pl = GFX_WRITE_ADDR(SLIST);

        pw = (ushort *) GFXDATA;

        for (i= 0 ; i < nbr ; i++)
                *pw = 0x00;
}


/***************** SET_COLORS *******************
 * Set the color in the palette number (0-3).   *
 *                                              *
 *   Made by P.Lee                              *
 *   Modified by Kaneda to use arg and any pal  *
 ************************************************/
void Gen_SetPal(uchar number,ushort *pal)
{
	register uint *pl;
  register volatile ushort *pw;
  ushort decal;
  uchar x;

	/* Pal 1 is 32bytes, 16colors of 2bytes each, after Pal 0 */
  decal = number * 32;

	/* Point on adress memory of the pal we want to modif/create */
	pl = (uint *) GFXCNTL;
	*pl = GFX_COLOR_WRITE_ADDR(decal);

	pw = (ushort *) GFXDATA;
	for (x = 0; x < 16; x++)
		*pw=pal[x];
}

/********************** FADE_IN *************************
 * Fade in the palette PALNUM to the new palette PALDEST*
 *                                                      *
 * Made by P.Lee                                        *
 * Mod. by Kaneda                                       *
 *   fade to something else than the black pal          *
 ********************************************************/
void fade_in(palDest, palNum)
ushort *palDest;
uchar palNum;
{
    ushort tmp_pal[16];
    register uint *pl;
    register ushort *pw;
    register ushort *pal_ptr;
    register ushort i, j, c;
    register uint Decal;

    /* see set_colors() */
    Decal = (32 * palNum);

    /* Initialize a temporary palette */
    for (i = 0; i < 16; i++)
        tmp_pal[i] = 0;

    /* Fade in loop */
    for (i = 0; i < 16; i++)
    {
        /* Pointer to palette we want */
        pal_ptr = palDest;
    
        /* Setup GFX to save colors */
        pl = (uint *) GFXCNTL;
        *pl = GFX_COLOR_WRITE_ADDR(Decal);

        /* Setup GFX data register */
        pw = (ushort *) GFXDATA;
        
        /* Increment one color B,G,R at a time */
        for (j = 0; j < 16; j++)
        {
            /* increment BLUE till reaching target color */
            c = tmp_pal[j] & 0xf00;
            if (c < (*pal_ptr & 0xf00))
                tmp_pal[j] += 0x100;
            /* check GREEN */
            c = tmp_pal[j] & 0xf0;
            if (c < (*pal_ptr & 0xf0))
                tmp_pal[j] += 0x10;
            /* check RED */
            c = tmp_pal[j] & 0xf;
            if (c < (*pal_ptr++ & 0xf))
                tmp_pal[j] += 0x1;

            /* Store color */
            *pw = tmp_pal[j];
        }

        /* Wait a sec... */
        for (j = 0; j < 4; j++)
            wait_sync();
    }
}


/********************* INIT_JOYPADS *************
 *  Initialize Joypads                          *
 *                                              *
 *   Made by P.Lee                              *
 ************************************************/
void init_joypad()
{
    register volatile uchar *pb;

    pb = (uchar *) 0xa10009;
    *pb = 0x40;
    pb += 2;
    *pb = 0x40;
    pb += 2;
    *pb = 0x40;
}

/*********************** READ_JOYPAD1 *******************
 *  Read joypad #1                                      *
 *  Return a byte like this :                           *
 *                              76543210                *
 *                              SACBRLDU                *
 *  A = button A                                        *
 *  B = button B                                        *
 *  C = button C                                        *
 *  D = Down                                            *
 *  L = Left                                            *
 *  R = Right                                           *
 *  U = Up                                              *
 *  S = Start                                           *
 *                                                      *
 *  Made by P.Lee                                       *
 *      to be modified for 6 buttons use                *
*********************************************************/
ushort read_joypad1()
{
    register volatile uchar *pb;
    ushort i, j;

    pb = (uchar *) 0xa10003;

    *pb = 0x40;        /* check joypad */
    asm("nop");
    asm("nop");
    i = *pb & 0x3f;

    *pb = 0;           /* check buttons */
    asm("nop");
    asm("nop");
    j = (*pb & 0x30) << 2;

    return( ~(i|j) );
}

/************************** READ_JOYPAD2 ****************
 * see read_joypad1 for more details                    *
 ********************************************************/
ushort read_joypad2()
{
    register volatile uchar *pb;
    ushort i, j;

    pb = (uchar *) 0xa10005;

    *pb = 0x40;        /* check joypad */
    asm("nop");
    asm("nop");
    i = *pb & 0x3f;

    *pb = 0;           /* check buttons */
    asm("nop");
    asm("nop");
    j = (*pb & 0x30) << 2;

    return(~(i | j));
}


/********************** WAIT_PRESS **********************
 *  Wait for a button to be pressed                     *
 *                                                      *
 *  Button can be A, B, C or Start not direction!!      *
 *                                                      *
 *  Made by Kaneda                                      *
 ********************************************************/
void wait_press()
{
    uchar tag=0;
    ushort i;

    init_joypad();
    while(tag == 0)
    {
        i=read_joypad1();

        if (i)
        {
                switch(i&0xff)
                {
                        case BUTTON_A:
                        case BUTTON_B:
                        case BUTTON_C:
                        case BUTTON_S:
                                tag=1;
                                break;
                }
        }
    }
}

/*********************** SCROLLV_A **********************
 * Scroll the mapfield A                                *
 * pos < 0 , scroll down                                *
 * pos > 0 , scroll up                                  *
 *                                                      *
 * Made by P.Lee                                        *
 * Modified by Kaneda                                   *
 ********************************************************/
void scrollv(posA, posB)
short int posA;
short int posB;
{
    register volatile short int *pw;
    register uint *pl;

    /* Setup GFX pointers */
    pl = (uint *) GFXCNTL;
    *pl = GFX_VERT_SCROLL(0); /* 0 if all scroll else, the number of col */

    pw = (short int *) GFXDATA;
    *pw = posA;
    *pw = posB;
}




/************************* SCROLLH **********************
 * Scroll mapfield A and B                              *
 * pos < 0 , scroll left                                *
 * pos > 0 , scroll right                               *
 *                                                      *
 * Made by Kaneda                                       *
 * 0.1 Delete the wait_sync()				*
 ********************************************************/
void scrollh(posA, posB)
short int posA;
short int posB;
{
    register volatile short int *pw;
    register ulong *pl;

    pl = (ulong *) GFXCNTL;
    *pl = GFX_WRITE_ADDR(HSCRL);

    pw = (short int *) GFXDATA;
    *pw = posA;
    *pw = posB;
    /* you must scroll fieldA before fieldB....(?) */
}



/*********************** RAZ ****************************
 * RAZ of fields, sprites, palettes, scrolls            *
 *                                                      *
 *  Made by Kaneda                                      *
 *	0.1 : delete wait sync call			*
 ********************************************************/
void RAZ()
{
	uchar i,j;
	
   Gen_SetPal(0, KANGEN_pal_black);
   Gen_SetPal(1, KANGEN_pal_black);
   Gen_SetPal(2, KANGEN_pal_black);
   Gen_SetPal(3, KANGEN_pal_black);

   Gen_ResetSprites(80);

   scrollv(0,0);
   scrollh(0,0);
   
   
   for (j=0;j<extHEIGHT;j++)
	{
		for( i=0; i<extWIDTH; i++)
		{
			show_tiles(0, 0, PAL0, i, j, BPLAN, 0, 0, 0);
			show_tiles(0, 0, PAL0, i, j, APLAN, 0, 0, 0);
		}
	}
}

/************************** WAIT_SYNC *******************
 *  Wait for VBL.                                       *
 *                                                      *
 *  Made by P.Lee                                       *
 ********************************************************/
void wait_sync()
{
    uint x;

    x = vtimer;
    while (x == vtimer);
    /*
    if ( (vtimer-x) > 1 )
    {
    	AlertEmu("Aie!");
    	AlertNumber( (vtimer-x) );
    }
    */
}
