#ifndef _INC_GENESIS_
#define _INC_GENESIS_

/*
 * genesis.h
 *
 * Copyright (c) 1996 by Paul W. Lee
 *
 * Permission is granted to anyone to use this software for any purpose
 * on any computer system, and to redistribute it freely, with the
 * following restrictions:
 * 1) No charge may be made other than reasonable charges for reproduction.
 * 2) Modified versions must be clearly marked as such.
 * 3) The authors are not responsible for any harmful consequences
 *    of using this software, even if they result from defects in it.
 */

/*
 *      GFX Registers
 */
#define GFXDATA 0xc00000
#define GFXCNTL 0xc00004

#define Z80_HALT  0xa11100
#define Z80_RESET 0xa11200
/*
 *      Joypad definitions
 */
#define JOY_UP          1
#define JOY_DOWN        2
#define JOY_LEFT        4
#define JOY_RIGHT       8
#define BUTTON_A        0x40
#define BUTTON_B        0x10
#define BUTTON_C        0x20
#define BUTTON_S        0x80

/*
 *      GFX address macros
 */
#define GFX_WRITE_ADDR(x) ((0x4000 + (x & 0x3fff)) << 16) + (x >> 14)
#define GFX_DMA_ADDR(x)   ((0x4000 + (x & 0x3fff)) << 16) + ((x >> 14) | 0x80)
#define GFX_COLOR_WRITE_ADDR(x) ((0xc000 + (x & 0x3fff)) << 16) + (x >> 14)
#define GFX_COLOR_READ_ADDR(x)  ((0x0000 + (x & 0x007f)) << 16) + (x >> 14) + 0x0020

#define GFX_VERT_SCROLL(x) ((0x4000 + (x & 0x3fff)) << 16) + ((x >> 14) | 0x10)
#define GFX_HORZ_SCROLL(x) ((0x4000 + (x & 0x3fff)) << 16) + (x >> 14)   

/*
 *  Field Adr define
 */

#define WPLAN 0xF000 /* mask : 0xF000 | Depend of window size...for a 32x32, 2048 max / 0x800 */
#define APLAN 0xC000 /* mask : 0xE000 | in 64*64, 4096b / 0x1000 */
#define BPLAN 0xE000 /* mask : 0xE000 | in 64*64, 4096b / 0x1000 */
#define SLIST 0xBE00 /* mask : 0xFC00 | max size 640b / 0x280 => 80 sprites */
#define HSCRL 0xB800 /* mask : 0xFC00 | B800 - 0000 = B800 /32 = 1472 tiles max | size max 1024b / 0x400 */


#define MIN(x,y)	((x>y)?y:x)
#define MAX(x,y)	((x>y)?x:y)
#define ABS(x)		((x<0)?-x:x)


/*
 *  Modify this else it could bug :)
 *      WIDTH  = Width of the screen mod by reg #16
 *      HEIGHT = Height of the screen mod by reg #16
 *
 */

#define HEIGHT  32


/*
 *      Misc defines
 */
#define FALSE   0
#define TRUE    1
//#define NULL    0

#define	PAL0	0
#define	PAL1	1
#define	PAL2	2
#define	PAL3	3

#define	NO_FLIP	0
#define	FLIPV	1
#define FLIPH	2

#define BOOL		unsigned char
#define uchar   unsigned char
#define ushort	unsigned short int
#define uint    unsigned int
#define ulong   unsigned long

struct TileData
{
	ushort compressSize;
	ushort width;
	ushort height;
	ushort nbTiles; // or idxTiles for sprites...
	uint   *data;
};
#define TILEDATA struct TileData
#define SPRDATA struct TileData

volatile uint 	vtimer;            /* VBL timer */
volatile uchar 	ResetState;
volatile uchar 	PauseState;

/******* Function prototypes in UTILS.C ************/
void Gen_ScreenSize( ushort width, ushort height);

void dma_vram_copy(uint source, ushort dest, ushort len);
void show_tiles(ushort from, ushort to, uchar pal, ushort posx, ushort posy, ushort fieldAdr, uchar pri, uchar flipV, uchar flipH);
void Gen_ShowUnique(ushort posx, ushort posy, ushort tileDef, ushort fieldAdr);//, uchar pal);

BOOL Gen_DefineSprite(uchar number,ushort posx,ushort posy,ushort size,ushort from);
void Gen_ShowSprites(uchar from,uchar to);
void Gen_ResetSprites(uchar nbr);

void Gen_SetPal(uchar number,ushort *pal);
void fade_in(ushort *palDest,uchar palNum);

void init_joypad();
ushort read_joypad1();
ushort read_joypad2();

void scrollv_A(short int pos);
void scrollv(short int posA,short int posB);
void scrollh(short int posA,short int posB);

void wait_press();
void wait_sync();
void RAZ();

#endif
