#include "genesis.h"
#include "anim.h"
#include "collide.h"

#include "debug.h"


uchar collideData[32][30]; 

/***** COLLIDE_SPRITE ******
 * check if 2 animSprites collide
 ****************************/
BOOL Collide_Sprite( uchar spriteRef, uchar spriteCmp )
{
	volatile ushort a, b;
	struct str_Frame *frameRef, *frameCmp;
	
	if (Anim[ spriteCmp ].IDCode == FONT)	return FALSE;
	if (Anim[ spriteCmp ].IDCode == EXPLOSION)	return FALSE;
	
	if ((Anim[ spriteRef ].IDCode&0x10) && (Anim[ spriteCmp ].IDCode&0x10))	return FALSE; //don't check shot VS shot

	frameRef = (struct str_Frame *) Anim[spriteRef].AnimLib->Frames[ Anim[spriteRef].curFrame ];
	frameCmp = (struct str_Frame *) Anim[spriteCmp].AnimLib->Frames[ Anim[spriteCmp].curFrame ];
	
	// first, Y Collide, more often
	a = Anim[spriteCmp].y + frameCmp->CBoxes[0].Y;
	b = Anim[spriteRef].y + frameRef->CBoxes[0].Y + frameRef->CBoxes[0].Height;
	if ( a > b )	return FALSE;

	a = Anim[spriteRef].y + frameRef->CBoxes[0].Y;
	b = Anim[spriteCmp].y + frameCmp->CBoxes[0].Y + frameCmp->CBoxes[0].Height;
	if ( a > b )	return FALSE;

	a = Anim[spriteCmp].x + frameCmp->CBoxes[0].X;
	b = Anim[spriteRef].x + frameRef->CBoxes[0].X + frameRef->CBoxes[0].Width;
	if ( a > b )	return FALSE;


	a = Anim[spriteRef].x + frameRef->CBoxes[0].X;
	b = Anim[spriteCmp].x + frameCmp->CBoxes[0].X + frameCmp->CBoxes[0].Width; ;
	if ( a > b )	return FALSE;


	Script_Kill( spriteRef, spriteCmp );
	return TRUE;
}


/***** COLLIDE_UPDATE ******
 * update animSprite collision data
 ****************************/
void Collide_Update( uchar spriteRef )
{		
	uchar i, j;
	uchar firstCol, lastCol;
	uchar firstLin, lastLin;
	ushort posx, posy;
	struct str_Frame *frame;

	
	if 	(Anim[ spriteRef ].active == FALSE)	return;
	if	(Anim[ spriteRef ].IDCode&0x80)	return; //only save enemy data
	if 	(Anim[ spriteRef ].x < 128)	return;
	if 	(Anim[ spriteRef ].x > 384)	return; //(128+ 32*8)
		
	frame = (struct str_Frame *) Anim[spriteRef].AnimLib->Frames[ Anim[spriteRef].curFrame ];

	posx = Anim[ spriteRef ].x + frame->CBoxes[0].X - 128;
	firstCol =  (uchar) (posx / 8);
//	firstCol = MAX(0, firstCol);
	
	lastCol = (posx + frame->CBoxes[0].Width)/8;
	lastCol = MIN(31, lastCol);
	
	posy = Anim[ spriteRef ].y + frame->CBoxes[0].Y - 128;
	firstLin =  (uchar) posy / 8;
//	firstLin = MAX(0, firstLin);
	
	lastLin = (posy + frame->CBoxes[0].Height)/8;
	lastLin = MIN(29, lastLin);
		
	spriteRef++; // 1 based, because 0 is "no sprite"
	for (j=firstLin; j<= lastLin; j++)
	{
		for (i=firstCol; i<= lastCol; i++)
		{
			collideData[i][j] = spriteRef;
		}
	}
}

/***** COLLIDE_CHECK *******
 * check animSprite COULD collide
 * with others animsprites
 ****************************/
void Collide_Check( uchar spriteRef )
{
	uchar i, j;
	uchar firstCol, lastCol;
	uchar firstLin, lastLin;
	ushort posx, posy;
	uchar tmp;
	struct str_Frame *frame;
	
	if 	(Anim[ spriteRef ].active == FALSE)	return;
	if	(Anim[ spriteRef ].IDCode < 0x80)	return; //only check hero data
	if 	(Anim[ spriteRef ].x < 128)	return;
	if 	(Anim[ spriteRef ].x > 384)	return; //(128+ 32*8)
		
	frame = (struct str_Frame *) Anim[spriteRef].AnimLib->Frames[ Anim[spriteRef].curFrame ];

	posx = Anim[ spriteRef ].x + frame->CBoxes[0].X - 128;
	firstCol =  (uchar) (posx / 8);
//	firstCol = MAX(0, firstCol);
	
	lastCol = (posx + frame->CBoxes[0].Width)/8;
	lastCol = MIN(31, lastCol);
	
	posy = Anim[ spriteRef ].y + frame->CBoxes[0].Y - 128;
	firstLin =  (uchar) posy / 8;
//	firstLin = MAX(0, firstLin);
	
	lastLin = (posy + frame->CBoxes[0].Height)/8;
	lastLin = MIN(29, lastLin);
		
	/* check only the rows where this sprite move */
	for (j=firstLin; j<= lastLin; j++)
	{
		for (i=firstCol; i<= lastCol; i++)
		{
			tmp = collideData[i][j];
			if ( tmp )	Collide_Sprite( --tmp, spriteRef );
		}
	}			
}