#include "genesis.h"
#include "fm.h"

extern uchar 	music[];
extern uchar	bank[];


/*********** TO DO
 *
 * - rename function as FM_xxxx
 * - pitch (if possible)
 * - user voice
 * - ILoop
 * - optimize FreqLow (too much duplicate value) ?
 * - PSG
 * - channel3/6
 * - PCM
 *
 ************************/
#define FM1_REG   0xa04000
#define FM1_DAT   0xa04001
#define FM2_REG   0xa04002
#define FM2_DAT   0xa04003

struct MMDHeader{
	ushort	TrackA;
	ushort	TrackB;
	ushort	TrackC;
	ushort	TrackD;
	ushort	TrackE;
	ushort	TrackF;
	ushort	TrackG;
	ushort	TrackH;
	ushort	TrackI;
	ushort	TrackJ;
	uchar	reserved;
	uchar	TimerB;
	ushort	UserVoices;
	uchar	*Title;
}*MMD;

struct MMDTrack
{	
	ushort	*Data;
	ushort	*Loop_Start;
	uchar	Loop_Count;
	uchar	Count;
	uchar	Note;
};

struct MMDVoice
{
	uchar	DET_MUL[4];
	uchar	TOT[4];
	uchar	RS_AR[4];
	uchar	D1R[4];
	uchar	D2R[4];
	uchar	D1L_RR[4];
	uchar	FB_ALG;
	uchar	name[7];	
};

volatile uchar vflag;

uchar FreqHigh[97]= {
	00,
	02,02,02,03,03,03,03,03,04,04,04,04,
	10,10,10,11,11,11,11,11,12,12,12,12,
	18,18,18,19,19,19,19,19,20,20,20,20,
	26,26,26,27,27,27,27,27,28,28,28,28,
	34,34,34,35,35,35,35,35,36,36,36,36,
	42,42,42,43,43,43,43,43,44,44,44,44,
	50,50,50,51,51,51,51,51,52,52,52,52,
	58,58,58,59,59,59,59,59,60,60,60,60
};

uchar FreqLow[97]={
	000,
	141,180,221,9,55,104,156,211,13,75,140,209,
	141,180,221,9,55,104,156,211,13,75,140,209,
	141,180,221,9,55,104,156,211,13,75,140,209,
	141,180,221,9,55,104,156,211,13,75,140,209,
	141,180,221,9,55,104,156,211,13,75,140,209,
	141,180,221,9,55,104,156,211,13,75,140,209,
	141,180,221,9,55,104,156,211,13,75,140,209,
	141,180,221,9,55,104,156,211,13,75,140,209
};


struct MMDTrack	Tracks[6];
struct MMDVoice	*Voices;

/*****************************************/
static void Z80_Pause( )
{
	register ushort *pw;

  pw = (ushort *) Z80_HALT;
	*pw = 0x100;	
}

static void Z80_Release( )
{	
	register ushort *pw;

	pw = (ushort *) Z80_HALT;
	*pw = 0;
}

/*****************************************/
static void FM_SetOP( uchar chan, uchar reg, uchar dat)
{
   	volatile register uchar *pFM;

   	if (chan < 3)
   	{
		pFM = (uchar *) FM1_REG;
		if ( (reg >= 0x30) && (reg <= 0xb4) )
			reg += chan;
   	}
   	else
   	{
		pFM = (uchar *) FM2_REG;
		if ( (reg >= 0x30) && (reg <= 0xb4) )
    		reg += (chan-3);
   	}

   	/* wait until "chip not busy" */
   	while (*pFM & 0x80);
   
   	*pFM = reg;   
   	pFM++;
   
   	while (*pFM & 0x80);

   	*pFM = dat;
   
   	while (*pFM & 0x80);
}

static void FM_ActiveNote( uchar chan, BOOL on )
{
	if (chan>2)	chan++;
	
	if (on)
		FM_SetOP( 0, 0x28, 0xF0 | chan);	
	else
		FM_SetOP( 0, 0x28, chan);	
}

static void FM_PlayNote( uchar chan, uchar note )
{
	FM_ActiveNote( chan, FALSE );
	
	FM_SetOP( chan, 0xA4, FreqHigh[note]);
	FM_SetOP( chan, 0xA0, FreqLow[note]);
	
	FM_ActiveNote( chan, TRUE );
}


static void FM_SetVoice( uchar chan, uchar idx, uchar vol )
{
	uchar	op;
	struct MMDVoice	*curV;
	
	curV = Voices;
	curV += idx;

	FM_ActiveNote( chan, FALSE );

	Tracks[chan].Note = idx;

	for( op = 0; op<4; op++)
	{
		FM_SetOP(chan, 0x30+(op*4), curV->DET_MUL[op]	);
		FM_SetOP(chan, 0x40+(op*4), curV->TOT[op]		);
   		FM_SetOP(chan, 0x50+(op*4), curV->RS_AR[op]	);
   		FM_SetOP(chan, 0x60+(op*4), curV->D1R[op]		);
		FM_SetOP(chan, 0x70+(op*4), curV->D2R[op]		);
		FM_SetOP(chan, 0x80+(op*4), curV->D1L_RR[op]	);
	}			
	
	/* handle volume */
	switch(curV->FB_ALG & 0x7)
	{
		case 7:
			FM_SetOP(chan, 0x40, ((curV->TOT[0])+vol)&0x7F	);				
		case 5:
		case 6:
			FM_SetOP(chan, 0x44, ((curV->TOT[1])+vol)&0x7F	);		
		case 4:
			FM_SetOP(chan, 0x48, ((curV->TOT[2])+vol)&0x7F	);
		case 0:
		case 1:
		case 2:
		case 3:
			FM_SetOP(chan, 0x4C, ((curV->TOT[3])+vol)&0x7F	);
			break;
	}
		
	FM_SetOP(chan,0xb0, curV->FB_ALG);
}

static void FM_SetVolume( uchar chan, uchar vol )
{
	FM_SetVoice( chan, Tracks[chan].Note, vol );	
}

static void FM_SetPan( uchar chan, uchar code )
{
	FM_SetOP(chan,0xb4, code);	
}

static void FM_SetTimerB( uchar val )
{
	FM_SetOP(0,0x26, val);	
	FM_SetOP(0,0x27, 0x2A);	//reset timer
}

static uchar FM_WaitOverFlow( )
{
   	volatile register uchar *pFM;

	pFM = (uchar *) FM1_REG;	
	return ((*pFM & 0x02));
}

static uchar FM_ReadMusic( uchar chan, ushort *command)
{
	uchar code;
	uchar val;

//	if (command == 0)	return 1;

	val = ((*command)&0xFF);
	code = ((*command)>>8)&0xFF;
	
	
	if ((code >= 1) && (code<=0x61))
	{
		//AlertEmu("Note");
		FM_PlayNote( chan, code );
		return	(val);
	}
		
	switch( code )
	{
		case 0x80:
			//AlertEmu("80 - Wait");
			FM_ActiveNote( chan, FALSE );
			return (val);
		
		case 0xF2:
			//AlertEmu("F2 - PitchDown");
			return (0);
		
		case 0xF4:
			//AlertEmu("F4 - BeginILoop");
			return (0);
			
		case 0xF6:
			//AlertEmu("F6 - PitchUp");		
			return (0);
 			
		case 0xF7:
			//AlertEmu("F7 - FM_SetVolume");
			FM_SetVolume( chan, val);
			return (0);
			
		case 0xF8:
			//AlertEmu("F8 - Stereo");
			FM_SetPan( chan, val);
			return (0);
			
		case 0xF9:
			//AlertEmu("F9 - Hold");		
			return (val);
			
		case 0xFA:
			//AlertEmu("FA - LoopEnd");
			if (val == 1)	return(0); /* simply a note list */
			
			if (val == 0)
			{
				/* infinite loop */
				Tracks[chan].Data = Tracks[chan].Loop_Start;
				return(0);
			}
			
			/* if no current loop, create one */
			if (Tracks[chan].Loop_Count == 1)
				Tracks[chan].Loop_Count = val;

			/* one count */
			Tracks[chan].Loop_Count--;
			
			/* finish or loop ?*/
			if (Tracks[chan].Loop_Count != 1)
				Tracks[chan].Data = Tracks[chan].Loop_Start;											
			return (0);
			
		case 0xFB:
			//AlertEmu("FB - LoopStart");	
			Tracks[chan].Loop_Start = Tracks[chan].Data;				
			return (0);
			
		case 0xFC:
			//AlertEmu("FC - SetTimerB");	
			FM_SetTimerB( val );
			return (0);
			
		case 0xFD:
			//AlertEmu("FD - SetUserVoice");		
			return (0);
 			
		case 0xFE:		
			//AlertEmu("FE - End/EndILoop");		
			return (1);
			
		case 0xFF:
			//AlertEmu("FF - FM_SetVoice");
			FM_SetVoice( chan, val, 0);
			return (0);
			
		default:
			//AlertEmu("FM Error");
			//AlertNumber(code);
			return (1); //skip
			
	}	
}

static void FM_InitYM( uchar timerB )
{	
	uchar i;

	Z80_Pause( );

	FM_SetOP( 0, 0x22, 0);
	FM_SetOP( 0, 0x24, 0);
	FM_SetOP( 0, 0x25, 0);
	FM_SetTimerB( timerB );	
	FM_SetOP( 0, 0x28, 0);
	FM_SetOP( 0, 0x28, 1);
	FM_SetOP( 0, 0x28, 2);
	FM_SetOP( 0, 0x28, 4);
	FM_SetOP( 0, 0x28, 5);
	FM_SetOP( 0, 0x28, 6);

	FM_SetOP( 0, 0x2A, 0);	
	FM_SetOP( 0, 0x2B, 0);	

	FM_SetOP( 0, 0xB6, 0xC0);		


	Z80_Release( );	
}

/*******************************************************/
void FM_Init( )
{
	MMD = (struct MMDHeader	*)	music;
	Voices = (struct MMDVoice	*)	bank;
	
	Tracks[0].Data = (ushort *) ( music + MMD->TrackA);
	Tracks[0].Loop_Start = 0;
	Tracks[0].Loop_Count = 1;
	Tracks[0].Count = 0;
	Tracks[0].Note = 0;

	Tracks[1].Data = (ushort *) ( music + MMD->TrackB);
	Tracks[1].Loop_Start = 0;
	Tracks[1].Loop_Count = 1;
	Tracks[1].Count = 0;
	Tracks[1].Note = 0;

	Tracks[2].Data = (ushort *) ( music + MMD->TrackC);
	Tracks[2].Loop_Start = 0;
	Tracks[2].Loop_Count = 1;
	Tracks[2].Count = 0;
	Tracks[2].Note = 0;
	
	Tracks[3].Data = (ushort *) ( music + MMD->TrackD);
	Tracks[3].Loop_Start = 0;
	Tracks[3].Loop_Count = 1;
	Tracks[3].Count = 0;
	Tracks[3].Note = 0;
	
	Tracks[4].Data = (ushort *) ( music + MMD->TrackE);
	Tracks[4].Loop_Start = 0;
	Tracks[4].Loop_Count = 1;
	Tracks[4].Count = 0;
	Tracks[4].Note = 0;
	
	Tracks[5].Data = (ushort *) ( music + MMD->TrackF);
	Tracks[5].Loop_Start = 0;
	Tracks[5].Loop_Count = 1;
	Tracks[5].Count = 0;
	Tracks[5].Note = 0;
	
	FM_InitYM( MMD->TimerB );
}

void FM_Close( )
{
	MMD = 0;
	FM_InitYM( 0 );
}


void FM_Update( )
{
	uchar track;
	
	if (MMD==0)	return;
	
	Z80_Pause( );
	
	if ( FM_WaitOverFlow( ) )
	{
		FM_SetTimerB( MMD->TimerB );
		for(track = 0; track<6; track++)
		{
			if (Tracks[track].Data)
			{
				while( Tracks[track].Count == 0)
				{			
					Tracks[track].Count = FM_ReadMusic(track, Tracks[track].Data );			
					Tracks[track].Data++;
				}
				
				Tracks[track].Count--;
				if (Tracks[track].Count == 0)
					FM_ActiveNote( track, FALSE );
			}
		}		
	}
	Z80_Release( );
}