#include "genesis.h"

#include "control.h"

#include "anim.h"
//#include "collide.h"
//#include "sprite.h"

#include "level.h"
//#include "scroll.h"
//#include "script.h"


#define VNOTHING	0x00 /* desactivate vflag */
#define VSTD			0x01 /* for wait sync only */
#define VGAME			0x02 /* during game */
#define VSOM			0x0F /* in main page */


volatile uchar vflag;
volatile uchar nState;

void AlertNumber(ulong nVal)
{
	uchar string[8];
	
	string[7] = 0;
	string[6] = (nVal&0x0F)+48;
	if (string[6]>57)	string[6]+=7;
	nVal >>=4;
	string[5] = (nVal&0x0F)+48;
	if (string[5]>57)	string[5]+=7;
	nVal >>=4;
	string[4] = (nVal&0x0F)+48;
	if (string[4]>57)	string[4]+=7;
	nVal >>=4;
	string[3] = (nVal&0x0F)+48;
	if (string[3]>57)	string[3]+=7;
	nVal >>=4;
	string[2] = (nVal&0x0F)+48;
	if (string[2]>57)	string[2]+=7;
	nVal >>=4;
	string[1] = (nVal&0x0F)+48;
	if (string[1]>57)	string[1]+=7;
	nVal >>=4;
	string[0] = (nVal&0x0F)+48;
	if (string[0]>57)	string[0]+=7;
	
	AlertEmu( string );
}


/*****************  Initialize GFX  ****************/
/* Read the comment for more info on each register */
/***************************************************/
void Init_GFX( )
{
    volatile register ushort *pw;
    pw = (ushort *) GFXCNTL;

    *pw = 0x8004;   /* reg. 0 - disEnable HBL */
    *pw = 0x8174;   /* reg. 1 - Enable display, VBL, DMA + VCell size (28) */
    *pw = 0x8230;   /* reg. 2 - Plane A =$30*$400=$C000 */
    *pw = 0x832C;   /* reg. 3 - Window  =$2C*$400=$B000 */
    *pw = 0x8407;   /* reg. 4 - Plane B =$7*$2000=$E000 */
    *pw = 0x855F;   /* reg. 5 - sprite table begins at $BE00=$5F*$200 */
    *pw = 0x8600;   /* reg. 6 - not used */
    *pw = 0x8700;   /* reg. 7 - Background Color number*/
    *pw = 0x8800;   /* reg. 8 - not used */
    *pw = 0x8900;   /* reg. 9 - not used */
    *pw = 0x8a00;   /* reg 10 - HInterrupt timing */
    *pw = 0x8b00;   /* reg 11 - $0000abcd a=extr.int b=vscr cd=hscr */
    *pw = 0x8c00;   /* reg 12 - hcell mode + shadow/highight + interlaced mode */
    *pw = 0x8d2E;   /* reg 13 - HScoll Table = $B800 */
    *pw = 0x8e00;   /* reg 14 - not used */
    *pw = 0x8f02;   /* reg 15 - auto increment data */
    *pw = 0x9010;   /* reg 16 - scrl screen v&h size */
    Gen_ScreenSize( 32, 64 );
   
    *pw = 0x9100;   /* reg 17 - window hpos */
    *pw = 0x9200;   /* reg 18 - window vpos */
};

/**/
void Int_Error( )
{
	AlertEmu("Interrupt error");	
}

void VBlank_sub( )
{
	vtimer++;
	if (nState == 0)	return;
	
//	AlertEmu("start");
	switch(vflag)
	{
		case VNOTHING:
			break;
		
		case VSTD:
			if (nState == 1)
			{
				nState = 0;
				Anim_Update1( );
				nState = 2;
			}
			else
			{
				nState = 0;
				Anim_Update2( );				
				Ctrl_Read( );
				nState = 1;
			}
			break;
			
		case VSOM:
			Ctrl_Read( );
			Anim_Update1( );
			Anim_Update2( );
			Level_Update( );
			break;
			
		case VGAME:
			if (nState == 1)
			{
				nState = 0;
				FM_Update( );
				Anim_Update1( );
				nState = 2;
			}
			else
			{
				nState = 0;
				Ctrl_Read( );
				Anim_Update2( );
				Level_Update( );
				nState = 1;
			}
			break;
	}

//	AlertEmu("end");
}

/**/

int main( )
{
	uchar	level;
	
	nState = 0;
	vflag = VNOTHING;

	Init_GFX( );

/********** intro *********/
	RAZ();	
	FM_Close( );
	Ctrl_Init( 0 );
	
	Level_Init( 0xFF, 1 );
	
	vflag = VSTD;
	nState = 1;
	while(Joypad1.key == 0)
	{    		  
		wait_sync( );
	}

/*********** game ************/
	level = 0;
	while(1)
	{
		nState = 0;
		vflag = VNOTHING;
		
		RAZ();		
		Anim_Init( );
		FM_Close( );
		Ctrl_Init( 0 );
		if (level==0)	Ctrl_Init( 10 );

		Level_Init( level, 1 );

		vflag = VGAME;
		if (level==0)	vflag = VSOM;
		nState = 1;
		while( curLevel.idx == level)
		{
			wait_sync( );
		}
		level = curLevel.idx;
	}	
	return (0);
}