#include "genesis.h"
#include "level.h"
#include "anim.h"
#include "script.h"
#include "control.h"
#include "debug.h"


/********* TO DO **********
 *
 * - BHV_POINT
 * - others bhv list
 *
 **************************/

#define	MAX_Y	( 128 + 32*8 + 4*8 )
#define MIN_Y	( 128 - 4*8 )

#define	MAX_X	( 128 + 32*8 + 4*8 )
#define MIN_X	( 128 - 4*8 )


#define	BHV_END		0 //...
#define	BHV_LINE	1 // follow a line (+speedX, +speedY)
#define	BHV_WAIT	2	// wait and do
#define	BHV_GO		3 // go (x+tileX, y+tileY)
#define	BHV_POINT 4 // go (destX, destY)

#define	ACT_NONE	0
#define	ACT_SHOT	1

struct BHV_List
{
	uchar		BHV_Code;
	char		BHV_Data1;
	char		BHV_Data2;
	char		BHV_Data3; //action on ending
};

extern struct	BHV_List	Full_Down[];
extern struct	BHV_List	Fire_Down[];
extern struct	BHV_List	Go_Diag1[];
extern struct	BHV_List	Go_Diag2[];
extern struct	BHV_List	Go_Wait[];


char mycos[36]={
	100,
	98,
	94,
	87,
	77,
	64,
	50,
	34,
	17,
	0,
	-17,
	-34,
	-50,
	-64,
	-77,
	-87,
	-94,
	-98,
	-100,
	-98,
	-94,
	-87,
	-77,
	-64,
	-50,
	-34,
	-17,
	0,
	17,
	34,
	50,
	64,
	77,
	87,
	94,
	98
};
	
char mysin[36]={
	0,
	17,
	34,
	50,
	64,
	77,
	87,
	94,
	98,
	100,
	98,
	94,
	87,
	77,
	64,
	50,
	34,
	17,
	0,
	-17,
	-34,
	-50,
	-64,
	-77,
	-87,
	-94,
	-98,
	-100,
	-98,
	-94,
	-87,
	-77,
	-64,
	-50,
	-34,
	-17
};

ushort mytan[9]={
	0,
	18,
	36,
	58,
	84,
	119,
	173,
	275,
	570
};
	


static uchar	counter; // ..before spawn a new sprite

static uchar Script_GetAngle( uchar idxSpr )
{
	uchar angle, i;
	short tan;
	
	
	if (Anim[heroSprite].x/8 == Anim[idxSpr].x/8)
		angle = 9;
	else	if (Anim[heroSprite].y/8 == Anim[idxSpr].y/8)
		angle = 0;
	else
	{
		angle = 9;
		tan = (100*((Anim[heroSprite].y/8)-(Anim[idxSpr].y/8)) / ((Anim[heroSprite].x/8) - (Anim[idxSpr].x/8)));
		if (tan < 0)	tan -= 2*tan;
		
		for( i = 0; i< 9; i++)
		{
			if ( mytan[i]< tan )	angle = i;
		}
	}

	if (Anim[heroSprite].x/8 <= Anim[idxSpr].x/8)	angle = 18-angle;		
	if (Anim[heroSprite].y/8 <= Anim[idxSpr].y/8)	angle = 36-angle;		
			
	return angle;
}


static void Script_Action( uchar idxSpr, uchar code )
{
	union sprProp	data;
	uchar angle;
	uchar tileX0;
	uchar tileY0;

	data.defVal = 0;
	
	
	switch(code)
	{
		case ACT_NONE:
			break;
						
		case ACT_SHOT:
			tileX0 = (Anim[idxSpr].x-128) /8;  //5 bits (0->31)
			tileY0 = (Anim[idxSpr].y-128) /8;  //5 bits (0->31)
			angle = Script_GetAngle( idxSpr );  //6 bits (0->35)
			
			data.shoot.cos = mycos[angle];
			data.shoot.sin = mysin[angle];
							
		 	Anim_New( SHOT, &shot_3_anim, shot3_idx, Anim[idxSpr].x+15, Anim[idxSpr].y+15, PAL2, data);
			break;
	}
}

/****** SCRIPT_NEW ******************
 * spawn a new sprite
 ************************************/
static void Script_New( uchar id, ushort xpos, ushort ypos)
{
	union sprProp	data;
	
	data.defVal = 0;
	
	switch( id )
	{
		case NMI_DOWN:
			data.defVal = 3; //quick
			Anim_New( NMI_DOWN, &npc_anim, npc_idx, xpos, ypos, PAL3, data);
			break;
			
		case NMI_V_V:
			Anim_New( NMI_V_V, &npc_anim, npc_idx, xpos, ypos, PAL3, data);
			break;

		case NMI_SHOT:
			data.defVal = 0;
			Anim_New( NMI_SHOT, &npc_anim, npc_idx, xpos, ypos, PAL3, data);
			break;

		case METEOR:
			data.defVal = 1;//slow
			Anim_New( METEOR, &block_anim, block_idx, xpos, ypos, PAL0, data);
			break;

		case METEOR_D1:
			data.defVal = 1;//slow
			Anim_New( METEOR_D1, &block_anim, block_idx, xpos, ypos, PAL0, data);
			break;

		case METEOR_D2:
			data.defVal = 1;//slow
			Anim_New( METEOR_D2, &block_anim, block_idx, xpos, ypos, PAL0, data);
			break;
	}
}

/****** SCRIPT_INIT ******************
 * init a script
 *************************************/
void Script_Init(  )
{
	uchar i;

	if ( curLevel.script == 0)	return;
	
	counter = curLevel.script[0];
	if (counter == 0xFF)	return;
	
	for (i=0; i< 36; i++)
	{
		while (counter==0)
		{
			//	Script_New( curLevel.script[1], 128+ curLevel.script[2]*8, 384- 8*i);
			// bypass the first sprites...it's for loop ;)
			
			curLevel.script+=3;
			
			counter = curLevel.script[0];
			if (counter == 0xFF)	return;
		}
		counter--;
	}	
}

/****** SCRIPT_LOOP ****************
 * restart the current script
 ***********************************/
void Script_Loop( )
{
	if ( curLevel.script == 0)	return;
	
	counter = curLevel.script[0];	
}


/****** SCRIPT_UPDATE ****************
 * update the current behaviors
 *************************************/
void Script_UpdateOne( uchar idxSpr )
{
	union sprProp	data;
	struct	BHV_List	*BHV_Cur;
	short	destX, destY;
	BOOL	bEnd;

	if (Anim[idxSpr].active == FALSE)	return;

	switch( Anim[idxSpr].IDCode )
	{
		case FONT:
			Anim[idxSpr].curFrame = Anim[idxSpr].data.defVal;
			return;
			break;
			
		case SELECTOR:
			return;
			break;
		
		case HERO:
			if (Joypad1.key&JOY_UP)
			{
				Anim[idxSpr].y -=	Anim[idxSpr].data.defVal;
				Anim[idxSpr].y =	MAX( Anim[idxSpr].y, 160);
			}
			else if (Joypad1.key&JOY_DOWN)
			{
				Anim[idxSpr].y +=	Anim[idxSpr].data.defVal;
				Anim[idxSpr].y =	MIN( Anim[idxSpr].y, 320);
			}
		
			if (Joypad1.key&JOY_LEFT)
			{
				Anim[idxSpr].x -=	Anim[idxSpr].data.defVal;
				Anim[idxSpr].x =	MAX( Anim[idxSpr].x, 128);
			}
			else if (Joypad1.key&JOY_RIGHT)
			{
				Anim[idxSpr].x +=	Anim[idxSpr].data.defVal;
				Anim[idxSpr].x =	MIN( Anim[idxSpr].x, 352);
			}

			if (Joypad1.key&BUTTON_A)			
				Anim_New( HERO_SHOT1, &shot_1_anim, shot1_idx, Anim[idxSpr].x+8, Anim[idxSpr].y - 16, PAL2, data);
			else if (Joypad1.key&BUTTON_B)
				Anim_New( HERO_SHOT2, &shot_2_anim, shot2_idx, Anim[idxSpr].x+7, Anim[idxSpr].y - 10, PAL2, data);

			if (Joypad1.key&BUTTON_C)
			{
				Anim[idxSpr].data.defVal++;
				if (Anim[idxSpr].data.defVal==5)
					Anim[idxSpr].data.defVal=2;
			}
				

			return;
			break;

		case HERO_SHOT1:
			Anim[idxSpr].y-=4;
			if( Anim[idxSpr].y < MIN_Y)	Anim[idxSpr].active = FALSE;
			return;
			break;

		case HERO_SHOT2:
			Anim[idxSpr].y-=7;
			if( Anim[idxSpr].y < MIN_Y)	Anim[idxSpr].active = FALSE;
			return;
			break;
			
		case METEOR:
		case NMI_DOWN:
			//AlertEmu("Full line");
			BHV_Cur = (struct	BHV_List	*)Full_Down;
			break;

		case NMI_V_V:
			//AlertEmu("GoWait");
			BHV_Cur = (struct	BHV_List	*)Go_Wait;
			break;

		case NMI_SHOT:
			//AlertEmu("Enemy shot");
			BHV_Cur = (struct	BHV_List	*)Fire_Down;
			break;

		case SHOT:
			Anim[idxSpr].x += (4*Anim[idxSpr].data.shoot.cos)/100;
			Anim[idxSpr].y += (4*Anim[idxSpr].data.shoot.sin)/100;
			
			if( Anim[idxSpr].y < MIN_Y)	Anim[idxSpr].active = FALSE;
			if( Anim[idxSpr].y > MAX_Y)	Anim[idxSpr].active = FALSE;

			if( Anim[idxSpr].x < MIN_X)	Anim[idxSpr].active = FALSE;
			if( Anim[idxSpr].x > MAX_X)	Anim[idxSpr].active = FALSE;
			return;
			break;

		case METEOR_D1:
			BHV_Cur = (struct	BHV_List	*)Go_Diag1;
			break;

		case METEOR_D2:
			BHV_Cur = (struct	BHV_List	*)Go_Diag2;
			break;
			
		case EXPLOSION:
			if ( Anim[idxSpr].curFrame = Anim[idxSpr].AnimLib->nbFrames-1)
			{
				Anim[idxSpr].active = FALSE;
			}
			return;
	}
	
	BHV_Cur += Anim[idxSpr].bhvIdx;
	bEnd=FALSE;
	switch( BHV_Cur->BHV_Code )
	{
		case BHV_END:
			//AlertEmu("->End");
			Anim[idxSpr].active = FALSE;
			break;
					
		case BHV_LINE:
			//AlertEmu("->Line");
			if ( Anim[idxSpr].data.defVal == 0)	Anim[idxSpr].data.defVal=1;

			if ( BHV_Cur->BHV_Data1 )
			{
				Anim[idxSpr].x += ( BHV_Cur->BHV_Data1 * Anim[idxSpr].data.defVal );
				if( Anim[idxSpr].x > MAX_X)	bEnd=TRUE;
				else if( Anim[idxSpr].x < MIN_X)	bEnd=TRUE;
			}

			if ( BHV_Cur->BHV_Data2 )
			{
				if (BHV_Cur->BHV_Data2<0)	Anim[idxSpr].flip = FLIPH;
				
				Anim[idxSpr].y += ( BHV_Cur->BHV_Data2 * Anim[idxSpr].data.defVal );
				if( Anim[idxSpr].y > MAX_Y)	bEnd=TRUE;
				else if( Anim[idxSpr].y < MIN_Y)	bEnd=TRUE;
			}
				
			break;

		case	BHV_WAIT:			
			if (Anim[idxSpr].data.wait.timer == 0)	Anim[idxSpr].data.wait.timer = (uchar) BHV_Cur->BHV_Data1;
			
			Anim[idxSpr].data.wait.timer--;
			if (Anim[idxSpr].data.wait.timer == 0)	bEnd=TRUE;
			break;
			
		case	BHV_GO:
			if ( Anim[idxSpr].data.defVal == 0 )
			{
				Anim[idxSpr].data.dest.tileX = BHV_Cur->BHV_Data1 + (Anim[idxSpr].x-128) / 8;
				Anim[idxSpr].data.dest.tileY = BHV_Cur->BHV_Data2 + (Anim[idxSpr].y-128) / 8;
			}
			destX = 8*Anim[idxSpr].data.dest.tileX + 128;
			destY = 8*Anim[idxSpr].data.dest.tileY + 128;
			
			Anim[idxSpr].flip = 0;
			
			if ( Anim[idxSpr].x == destX-1 )
				Anim[idxSpr].x++;
			else if ( Anim[idxSpr].x < destX )
				Anim[idxSpr].x+=2;
			else if ( Anim[idxSpr].x > destX )
			{
				Anim[idxSpr].flip |= FLIPV;
				Anim[idxSpr].x-=2;
			}

			if ( Anim[idxSpr].y == destY-1 )
				Anim[idxSpr].y++;
			else if ( Anim[idxSpr].y < destY )			
				Anim[idxSpr].y+=2;
			else if ( Anim[idxSpr].y > destY )
			{
				Anim[idxSpr].flip |= FLIPH;
				Anim[idxSpr].y-=2;
			}
						
			if ( (Anim[idxSpr].x == destX) && (Anim[idxSpr].y == destY) )	bEnd=TRUE;
			break;
	}
	
	if (bEnd==TRUE)
	{
		Script_Action( idxSpr, BHV_Cur->BHV_Data3);
		Anim[idxSpr].bhvIdx++;
		Anim[idxSpr].data.defVal = 0;
	}
}

/****** SCRIPT_STEP ****************
 * update the current script
 ***********************************/
void Script_Step(  )
{
	if ( curLevel.script == 0)	return;
	if (counter == 0xFF)	return;

	while (counter==0)
	{
		Script_New( curLevel.script[1], 128+ curLevel.script[2]*8, 96);
		
		curLevel.script+=3;
		
		counter = curLevel.script[0];
		if (counter == 0xFF)
		{
			curLevel.script -= (vtimer&0x0F) * 3; //return between 0 and 0xF script back
			counter = curLevel.script[0];
		}
		if (counter == 0xFF)		return; //just in case it go back of 0;
	}
	counter--;
}

/****** SCRIPT_KILL ****************
 * start the special Kill bhv
 *************************************/
void Script_Kill( uchar idGood, uchar idBad )
{
	// TO DO
	union sprProp	data;
	ushort x, y;
	short diffy;
	
	data.defVal = 0;

	switch( Anim[idGood].IDCode )
	{
		case HERO:
			curLevel.idx = 0xFE; //game over
			break;

		case HERO_SHOT1: //mega killer, only blocked by meteo (see below)
			//Anim[idGood].active = FALSE;
			break;

		case HERO_SHOT2:
			Anim[idGood].active = FALSE;
			break;

		case SHOT:
			Anim[idGood].active = FALSE;
			break;
	}			

	switch( Anim[idBad].IDCode )
	{
		case NMI_DOWN:
		case NMI_V_V:
		case NMI_SHOT:
			x=Anim[idBad].x;
			y=Anim[idBad].y;
			Anim[idBad].active = FALSE;
			
			diffy = Anim[heroSprite].y - y;
			heroScore += 10;
			if ( (diffy > 0) && (diffy < 64) )	heroScore += 10; // size nmi + limit = 32 + 32
			if ( (diffy > 0) && (diffy < 48) )	heroScore += 10; // size nmi + max limit = 32 + 16
			
			Anim_New( EXPLOSION, &explo_anim, explo_idx, x, y, PAL2, data);
			break;

		case METEOR:
		case METEOR_D1:
		case METEOR_D2:
			if ( Anim[idGood].IDCode == HERO_SHOT1)
				Anim[idGood].active = FALSE;
			
			//nothing, can't be destroyed BUT intercept shot2 :)		
			break;
	}		
}